; HSCSI v0.27
;
;
	.include IOCSCALL.MAC
	.include DOSCALL.MAC

	.cpu	68030

MPUTYPE		.equ	$00000CBC
EX_SPC		.equ	$00EA0000		; CZ-6BS1 SPC base address
IN_SPC		.equ	$00E96020		; Internal SPC base address
;
; SPC MBԸ52 registers
;
BDID		.equ	$01	; Bus Device ID
SCTL		.equ	$03	; SPC Control
SCMD		.equ	$05	; SPC Command
;nazo		.equ	$07	; nazo register
INTS		.equ	$09	; Interrupt Sense
PSNS		.equ	$0B	; Phase Sense
SDGC		.equ	$0B	; SPC Diagnostic Control
SSTS		.equ	$0D	; SPC status
SERR		.equ	$0F	; SPC Error Status
PCTL		.equ	$11	; Phase Control
MBC		.equ	$13	; Modified Byte Counter
DREG		.equ	$15	; Data Register
TEMP		.equ	$17	; Temporary Register
TCH		.equ	$19	; Transfer Counter High
TCM		.equ	$1B	; Transfer Counter Mid
TCL		.equ	$1D	; Transfer Counter Low

READ_OP_CODE	.equ	$28
WRITE_OP_CODE	.equ	$2A


	.text

HSCSI_top:
		.dc.l	-1
		.dc.w	$8000
		.dc.l	strategy
		.dc.l	interrupt
CheckName:	.dc.b	'HSCSI!**'

		.align	4
new_SCSI:
		.dc.b	$70
ExcludeID:	.ds.b	1
		btst.l	d4,d0
		beq	NoHSCSI
		moveq	#_S_READEXT,d0
		cmp.l	d0,d1
		beq	GoHSCSIrd
		moveq	#_S_WRITEEXT,d0
		cmp.l	d0,d1
		beq	GoHSCSIwr
		moveq	#_S_READ,d0
		cmp.l	d0,d1
		beq	GoHSCSIrd
		moveq	#_S_WRITE,d0
		cmp.l	d0,d1
		beq	GoHSCSIwr
NoHSCSI:	.dc.w	$4EF9			; jmp xxx.l
orig_SCSI:	.ds.l	1


HSCSI_save	.reg	d1/d3/d7/a1-a6

		.align	4
GoHSCSIwr:	movem.l	HSCSI_save,-(sp)
		moveq	#WRITE_OP_CODE,d7
		bra	@f

		.align	4
GoHSCSIrd:	movem.l	HSCSI_save,-(sp)
		moveq	#READ_OP_CODE,d7
@@:
		swap	d3
		clr.w	d3
		swap	d3
		beq	HSCSI_error		; transfer block == ZERO

		.dc.w	$4DF9			; lea xxx.l,a6
SPC_BASE:	.ds.l	1			; SPC base address
		lea	(PSNS,a6),a5
		lea	(INTS,a6),a4
		lea	(DREG,a6),a3
		lea	(SSTS,a6),a2
		lea	(SCMD,a6),a0

		bsr	Selection
		tst.l	d0
		bne	HSCSI_error

		bsr	CommandOut
		tst.l	d0
		beq	@f
		subq.l	#%100,d0
		beq	1f
		bra	HSCSI_error
@@:
		bsr	DataTransfer
		tst.l	d0
		bne	HSCSI_error
1:
		lea	(TEMP,a6),a3
		lea	(PCTL,a6),a2
		bsr	StatusMessage

;@@:		btst.b	#8,(a5)			; PSNS  busy?
;		bne	@b
;		move.b	(a4),(a4)		; INTS

HSCSI_exit:
		movem.l	(sp)+,HSCSI_save
		rts

HSCSI_error:	moveq	#-1,d0
		bra	HSCSI_exit

IllegalPhase:	moveq	#-1,d0
		rts

;
; SELECTION
;
		.align	4
Selection:
		moveq	#0,d0
		move.b	d0,(PCTL,a6)		; selection phase
@@:		move.b	(a2),d1			; SSTS
		andi.b	#%1111_0000,d1
		bne	@b			; ɂɂȂ܂ő҂
		bset.b	#4,(SCTL,a6)		; SCTL  set Arbitration
		.dc.b	$72			; moveq #xx,d1
HBDID:		.ds.b	1			; BDID  read SPC ID
		bset.l	d4,d1
		move.b	d1,(TEMP,a6)
		move.w	#3000,d1		; 
		movep.w	d1,(TCH,a6)
		moveq	#3,d1
		move.b	d1,(TCL,a6)
		move.b	(a4),(a4)		; INTS
		move.b	#%0010_0000,(a0)	; SCMD  Select command
@@:		move.b	(a4),d1			; INTS
		andi.b	#%0001_0100,d1
		beq	@b
		move.b	(a4),(a4)		; INTS
		cmpi.b	#%0001_0000,d1		; command complete
		bne	SelectionFail
		rts

SelectionFail:	move.b	d0,(TEMP,a6)
		moveq	#-1,d0
		rts

;
; STATUS IN & MESSAGE IN
;
		.align	4
StatusMessage:
		moveq	#0,d0
		moveq	#%111,d7
		moveq	#%1110_0100,d1
@@:		move.b	(a5),d3			; PSNS
		bpl	@b
		and.b	d7,d3			; execution phase
		cmpi.b	#%011,d3		; status phase
		bne	IllegalPhase
		bsr	GetBusByte
@@:		move.b	(a5),d3			; PSNS
		bpl	@b
		and.b	d7,d3			; execution phase
		cmp.b	d7,d3			; message phase
		bne	IllegalPhase
GetBusByte:	move.b	d3,(a2)			; PCTL
		move.b	d1,(a0)			; SCMD  set ACK/REQ
@@:		tst.b	(a5)			; PSNS
		bmi	@b
		move.b	(a3),d0			; TEMP
		move.b	#%1100_0100,(a0)	; SCMD  reset ACK/REQ
		swap	d0
		rts

;
; COMMAND OUT
;
		.align	4
CommandOut:
		move.l	d4,d1
		swap	d1
		lsl.b	#5,d1			; d1 = LUN
		rol.l	#8,d2
		rol.w	#8,d3
		moveq	#10,d0			; COMMAND length
		movep.l	d0,(TEMP,a6)
@@:		move.b	(a5),d0			; get phase
		bpl	@b
		andi.b	#%111,d0		; execution phase
		cmpi.b	#%010,d0		; command phase
		bne	IllegalPhase_cout
		move.b	d0,(PCTL,a6)
		move.b	(a4),(a4)		; INTS
		move.b	#%1000_0000,(a0)	; Transfer command
@@:		move.b	(a2),d0			; SSTS
		andi.b	#%1111_0000,d0
		cmpi.b	#%1011_0000,d0
		bne	@b
		moveq	#0,d0
@@		btst.b	#0,(a2)			; SSTS  DREG empty ??
		beq	@b

		move.b	d7,(a3)		; OP-code
		move.b	d1,(a3)		; LUN
		move.b	d2,(a3)		; LBA
		rol.l	#8,d2
		move.b	d2,(a3)		; LBA
		rol.l	#8,d2
		move.b	d2,(a3)		; LBA
		rol.l	#8,d2
		move.b	d2,(a3)		; LBA

@@:		btst.b	#0,(a2)			; SSTS
		beq	@b

		move.b	d0,(a3)		; zero
		move.b	d3,(a3)		; count
		rol.w	#8,d3
		move.b	d3,(a3)		; count
		move.b	d0,(a3)		; zero

@@:		move.b	(a4),d0			; INTS
		beq	@b
		move.b	d0,(a4)
		cmpi.b	#%0001_0000,d0
		bne	@f
		moveq	#0,d0
@@:		rts

IllegalPhase_cout:
		cmpi.b	#%011,d0		; status phase
		bne	@f
		moveq	#%100,d0
		rts
@@:		moveq	#-1,d0
		rts

;
; DATA TRANSFER
;
		.align	4
DataTransfer:
		move.l	d3,d1
		lsl.l	#8,d1
		lsl.l	d5,d1			; transfer length
		movep.l	d1,(TEMP,a6)
@@:		move.b	(a5),d1			; get phase
		bpl	@b			; REQ ?
		andi.b	#%111,d1		; Execution phase
		beq	@f			; data out phase
		cmpi.b	#%001,d1		; data in phase
		bne	IllegalPhase_dt
@@:		move.b	d1,(PCTL,a6)
		move.b	(a4),(a4)		; INTS
		move.b	#%1000_0000,(a0)	; Transfer command
@@:		move.b	(a2),d0			; SSTS
		andi.b	#%1111_0000,d0
		cmpi.b	#%1011_0000,d0
		bne	@b
		tst.b	d1
		bne	ReadTransfer
		bra	WriteTransfer

IllegalPhase_dt:
		moveq	#-1,d0
		cmpi.b	#%011,d1		; status phase
		bne	@f
		moveq	#0,d0			; Ƃ肠AIƂĂ
@@:		rts

		.align	4
ReadTransfer:
		movec	CACR,d7
		move.l	d7,d0
		bclr.l	#8,d0	; enable bit
		bset.l	#11,d0	; flush bit
		movec	d0,CACR			; DC flush & disable
		subq.w	#1,d3			; for dbra
DR_Wait:
		btst.b	#1,(a2)			; SSTS  DREG full ??
		beq	DR_Wait			; f[^ 8Bytes ҂
		moveq	#(512/32)-1,d1
Get_512:
		.rept	32
		move.b	(a3),(a1)+
		.endm
		dbra	d1,Get_512
		dbra	d3,DR_Wait
		movec	d7,CACR			; DC mode restore
TransExit:
@@:		move.b	(a4),d0			; INTS
		beq	@b
		move.b	d0,(a4)
		cmpi.b	#%0001_0000,d0
		bne	@f
		moveq	#0,d0
@@:		rts


		.align	4
WriteTransfer:
		lsl.l	#6,d3
DW_Wait:
		btst.b	#0,(a2)			; SSTS  DREG empty ??
		beq	DW_Wait			; ɂȂ܂ő҂
		.rept	8
		move.b	(a1)+,(a3)
		.endm
		subq.l	#1,d3
		bne	DW_Wait
		bra	TransExit
keeppr_bottom:


request:	.dc.l	0

strategy:	move.l	a5,request
		rts

interrupt:	movem.l	d1/a1/a5,-(sp)
		movea.l	(request,pc),a5
		move.b	(2,a5),d1
		bsr	device_initialize
		move.l	#device_bottom,(14,a5)
		move.b	d0,(4,a5)
		ror.w	#8,d0
		move.b	d0,(3,a5)
		movem.l	(sp)+,d1/a1/a5
		rts
device_bottom:


device_initialize:
		move.l	(18,a5),a1
@@:		tst.b	(a1)+
		bne	@b
1:		move.b	(a1)+,d0
		bne	2f
		move.b	(a1)+,d0
		beq	3f
2:		subi.b	#'0',d0
		bcs	1b
		cmpi.b	#7,d0
		bhi	1b
		bset.b	d0,(ExcludeID)
		bra	1b
3:
initialize:	pea	(DevCopyMsg,pc)
		DOS	_PRINT
		addq.l	#4,sp

		tst.b	d1
		bne	illegal_system

		bsr	check_system
		tst.l	d0
		beq	@f
		rts

@@:		pea	(new_SCSI,pc)
		move.w	#$01F5,-(sp)
		DOS	_INTVCS
		addq.l	#6,sp
		move.l	d0,(orig_SCSI)

		moveq	#0,d0
		move.b	(ExcludeID,pc),d0
		beq	@f
		movea.l	d0,a1
		pea	(hscsi_m,pc)
		DOS	_PRINT
		addq.l	#4,sp
		moveq	#0,d1
1:		move.l	a1,d0
		btst.l	d1,d0
		beq	2f
		pea	(id_m,pc)
		DOS	_PRINT
		move.w	d1,(sp)
		addi.w	#'0',(sp)
		DOS	_PUTCHAR
		addq.l	#4,sp
2:		addq.b	#1,d1
		cmpi.b	#7,d1
		bls	1b
		pea	(targ_m,pc)
		DOS	_PRINT
		addq.l	#4,sp

@@:		pea	(SetUpMsg,pc)
		DOS	_PRINT
		addq.l	#4,sp

		movec	CACR,d1
		ori.w	#%00_1000_0000_1000,d1
		movec	d1,CACR				; cache flush

		moveq	#0,d0
		rts


check_system:
		cmpi.b	#3,MPUTYPE
		bne	illegal_mpu
		lea	$00ed006f,a1
		cmpi.b	#$56,(a1)+
		bne	illegal_scsi
		move.l	#EX_SPC,d0
		btst.b	#3,(a1)
		bne	@f
		move.l	#IN_SPC,d0
@@:		move.l	d0,(SPC_BASE)
		movea.l	d0,a1
		move.b	(BDID,a1),(HBDID)
		moveq	#0,d0
		rts

illegal_mpu:	pea	(no_mpu_msg,pc)
		bra	@f
illegal_scsi:	pea	(no_scsi_msg,pc)
@@:		DOS	_PRINT
		addq.l	#4,sp
illegal_system:	move.w	#$0350,d0
		rts


command_exec:
		suba.l	a1,a1
		IOCS	_B_SUPER
		move.l	d0,d7

		tst.b	(a2)+
		beq	@f
1:		move.b	(a2)+,d0
		beq	@f
2:		subi.b	#'0',d0
		bcs	1b
		cmpi.b	#7,d0
		bhi	1b
		bset.b	d0,(ExcludeID)
		bra	1b

@@:		move.w	#$01F5,-(sp)
		DOS	_INTVCG
		addq.l	#2,sp
		movea.l	d0,a0
		cmpi.l	#'HSCS',(CheckName-new_SCSI,a0)
		bne	goKEEPPR
		cmpi.l	#'I!??',(CheckName-new_SCSI+4,a0)
		beq	goMFREE

goKEEPPR:	clr.b	d1
		bsr	initialize
		move.l	d0,d1
		movea.l	d7,a1
		IOCS	_B_SUPER
		tst.l	d1
		bne	error_exit
		move.w	#'??',(CheckName+6)
		clr.w	-(sp)
		pea	(keeppr_bottom-HSCSI_top).w
		DOS	_KEEPPR

goMFREE:
		movec	CACR,d1
		ori.w	#%00_1000_0000_1000,d1
		movec	d1,CACR				; cache flush

		movea.l	d7,a1
		IOCS	_B_SUPER
		pea	(DevCopyMsg+2,pc)
		DOS	_PRINT
		move.l	(orig_SCSI-new_SCSI,a0),-(sp)
		move.w	#$01F5,-(sp)
		DOS	_INTVCS
		pea	(HSCSI_top-new_SCSI-$f0,a0)
		DOS	_MFREE
		lea	(4+6+4,sp),sp
		tst.l	d0
		bne	@f

		pea	(ReSetMsg,pc)
		DOS	_PRINT
		addq.l	#4,sp
		DOS	_EXIT

@@:		pea	(mfreeFail,pc)
		DOS	_PRINT
		addq.l	#4,sp
error_exit:	move.w	#1,-(sp)
		DOS	_EXIT2


no_mpu_msg:	.dc.b	'68030pł',13,10,0
no_scsi_msg:	.dc.b	'SCSI܂',13,10,0
DevCopyMsg:	.dc.b	13,10,'X68030 HSCSI version 0.27 Copyright 1993,95 Ⴌ゠',13,10,0
SetUpMsg:	.dc.b	'HSCSIgݍ݂܂',13,10,0
ReSetMsg:	.dc.b	'HSCSI܂',13,10,0
mfreeFail:	.dc.b	'HSCSỈɎs܂',13,10,0
hscsi_m:	.dc.b	'HSCSI',0
id_m:		.dc.b	' ID:',0
targ_m:		.dc.b	' ANZX܂',13,10,0

	.end	command_exec
